import pandas as pd
import re
import numpy as np
from cem_utils import *
from normalization import normalize_signal_with_lim
import pysam
import pyslow5

def extract_feature(line, strand, base_shift=2, norm=True):
    global nucleotide_type
    read_id = line[0]
    # if read_id !='562eeb47-2b86-4fc7-abfc-5dce62f511ed':
    #     return None

    # tackle moves tag
    moves_string = line[14]
    moves_string = re.sub('ss:Z:', '', moves_string)
    moves_string = re.sub('D', 'D,', moves_string)
    moves_string = re.sub('I', 'I,', moves_string)
    # print(moves_string)
    moves = re.split(r',+', moves_string)
    moves = moves[:-1]
    # extract index and generate event_length and event_start
    insertion = 0
    event_length = []
    for i, item in enumerate(moves):
        if 'D' in item:
            deletion = int(item[:-1])
            for i in range(deletion):
                event_length.append(0)
        elif 'I' in item:
            if i == 0:
                continue
            else:
                return None
        elif '=' in item:
            return None
        else:
            event_length.append(int(item))
    # build event_length from move table
    read = s5.get_read(read_id, aux=["read_number", "start_mux"], pA=True)
    start_index = line[2]
    end_index = line[3]
    event_length = np.array(event_length)

    # identify RNA or DNA
    if nucleotide_type is None:
        if line[7] > line[8]:
            nucleotide_type = 'RNA'
        else:
            nucleotide_type = 'DNA'
    #  assert len_raw_signal in paf and blow5
    try:
        assert end_index - start_index == np.sum(event_length)
        assert read['len_raw_signal'] == line[1]
    except Exception:
        print("Warning: 1 read's length of signal is not equal between blow5 and paf")
        return None

    signal = read['signal']

    # create event start and flip all table
    signal = signal[start_index:end_index]
    if nucleotide_type == 'RNA':
        signal = np.flip(signal)
        event_length = np.flip(event_length)

    event_starts = event_length.cumsum()
    event_starts = np.insert(event_starts, 0, 0)[:-1]
    if norm:
        signal = normalize_signal_with_lim(signal)

    # index query and reference
    aligned_pair = info_dict[read_id]['pairs']
    qlen = info_dict[read_id]['query_length']
    # correct index about DNA and RNA
    if nucleotide_type == 'RNA':
        if strand == '+':
            gap = qlen - event_length.shape[0]
            aligned_pair[0] = aligned_pair[0] - gap
            aligned_pair = aligned_pair[aligned_pair[0] >= 0]
        elif strand == '-':
            aligned_pair[0] = event_length.shape[0] - aligned_pair[0] - 1
    else:
        if strand == '-':
            aligned_pair[0] = qlen - aligned_pair[0] - 1
        aligned_pair = aligned_pair[aligned_pair[0] < event_length.shape[0]]
    # base shift
    if nucleotide_type == 'RNA':
        aligned_pair[0] = aligned_pair[0].values + base_shift
        aligned_pair = aligned_pair[aligned_pair[0] < event_length.shape[0]]
    else:
        aligned_pair[0] = aligned_pair[0].values - base_shift
        aligned_pair = aligned_pair[aligned_pair[0] >= 0]

    if aligned_pair.shape[0] == 0:
        return None
    read_pos = aligned_pair[0].values
    ref_pos = aligned_pair[1].values

    # extract raw signal by event length and event start
    total_feature_per_reads = []
    try:
        raw_signal_every = [signal[event_starts[x]:event_starts[x] + event_length[x]] for x in
                            read_pos]
        # if aligned_pair.shape[0] < 11:
        #     draw_signal(signal[event_starts[read_pos[0]]:event_starts[read_pos[-1]+1]], event_starts[read_pos[0]:read_pos[-1]+1]-event_starts[read_pos[0]], base_list)
    except Exception:
        print(1)
    # calculate mean median and dwell time
    for i, element in enumerate(raw_signal_every):
        if event_length[read_pos[i]] == 0:
            continue
        temp = [read_id, np.mean(element), np.std(element), np.median(element), event_length[read_pos[i]], ref_pos[i]]
        total_feature_per_reads.append(temp)
    return total_feature_per_reads


def read_blow5(path, reference, pore, subsample_ratio=1, base_shift=2, norm=True,
               cpu=4, rna=True):
    global info_dict, s5, pbar
    slow5_file = path + ".blow5"
    fastq_file = path + ".fastq"
    identify_file_path(fastq_file)
    identify_file_path(slow5_file)
    paf_file = generate_paf_file(fastq_file, slow5_file, pore, rna)
    bam_file = generate_bam_file(fastq_file, reference, cpu)
    bam_file = pysam.AlignmentFile(bam_file, 'rb')

    s5 = pyslow5.Open(slow5_file, 'r')

    df = pd.read_csv(paf_file, sep='\t', header=None)
    df = pd.merge(df, info_df, how='inner', on=0)
    if df.shape[0] == 0:
        raise RuntimeError("cannot found the record from bam in your paf file. Please check your f5c command ... ")
    if df.shape[0] / info_df.shape[0] < 0.8:
        print('There are ' + str(info_df.shape[0] - df.shape[0]) + " reads not found in your paf file ...")
    pbar = tqdm(total=df.shape[0], position=0, leave=True)
    df["feature"] = df.apply(extract_feature, strand=strand, base_shift=base_shift, norm=norm, axis=1)
    pbar.close()

    df.dropna(inplace=True)
    num_aligned = df.shape[0]
    if subsample_ratio < 1:
        df = df.sample(frac=subsample_ratio)
    final_feature = []
    for item in df["feature"]:
        final_feature.extend(item)
    final_feature = pd.DataFrame(final_feature)
    final_feature.columns = ['Read ID', 'Mean', 'STD', 'Median', 'Dwell time', 'Position']

    final_feature['Position'] = final_feature['Position'].astype(int).astype(str)
    print('Extracted ', num_aligned, ' aligned reads from blow5 files')

    return final_feature, num_aligned, nucleotide_type

read_blow5('/t1/zhguo/Data/ASFV_RNA/ASFV_go/WGS1','/t1/zhguo/Data/ASFV_RNA/asfv.fa','r10')


import pandas as pd
import numpy as np
from nanoCEM.cem_utils import extract_kmer_feature
import plotnine as p9
from matplotlib import pyplot as plt

plt.rcParams['pdf.fonttype'] = 42

from statsmodels.multivariate.manova import MANOVA
from sklearn.decomposition import PCA
results_path='dna_sample/'
df = pd.read_csv(results_path+'/current_feature.csv')

# 创建PCA对象，并指定降维后的维度
# methylation_list = [
#     2503,1911, 1915,1917, 1939,2498,
#     1835,2552,2030,2604, 2261,746,745,744,1618,2445,
#     2069,1962,955,1594,2580,2457,2605,2501,
# ]
# df = df[df['Position']==2030]
# median_by_group = df.groupby('Group')['Mean'].median()
#
# print(median_by_group)

position=43281378
# position = 43281378

methylation_list=list(range( position- 9 ,position + 10))
i=0
result_list=[]
for item in methylation_list:
    # subsample the reads
    control = df[df['Group']=='Control']
    if control.shape[0] > 10500:
        control=control.iloc[0:10500, :]
    sample = df[df['Group'] == 'Sample']
    if sample.shape[0] > control.shape[0] * 2:
        sample = sample.iloc[0:control.shape[0],:]
    df = pd.concat([control,sample],axis=0).reset_index(drop=True)
    feature,label = extract_kmer_feature(df,3,item)
    y_test = label[0].apply(lambda x: 1 if x == 'Sample' else 0)
    #
    pca = PCA(n_components=2,whiten=True)
    new_df = pd.DataFrame(pca.fit_transform(feature))

    # reducer = umap.UMAP(n_components=2)  # 指定降维后的维度为2
    # new_df = umap.UMAP(
    #     n_neighbors=20,
    #     min_dist=0.0,
    #     n_components=2,
    # ).fit_transform(feature)
    new_df = pd.concat([pd.DataFrame(new_df),label], axis =1)
    new_df.columns=['PC1','PC2','Group']
    if np.sum(new_df['Group']=='Sample') > 10 and np.sum(new_df['Group']=='Control') > 10:
        manova = MANOVA.from_formula('PC1 + PC2 ~ Group', data=new_df)
        # 执行多元方差分析
        results = manova.mv_test()
        pvalue = results.summary().tables[3].iloc[0,5]
        result_list.append([item,pvalue])


new_df = pd.DataFrame(result_list)
new_df[1] = np.log10(new_df[1]) * (-1)
print(i)
new_df.columns=['Position','P value(-log10)']
new_df['Group'] = new_df['P value(-log10)'].apply(lambda x: 'Significant' if x>=2 else 'Not significant')
plot = p9.ggplot(new_df, p9.aes(x='Position', y='P value(-log10)',fill='Group'))\
    + p9.theme_bw() \
    + p9.geom_col() \
    + p9.scale_y_continuous(breaks=[2, 4, 6, 8, 10])\
    + p9.scale_fill_manual(values={'Significant': "#BB9A8B", 'Not significant': "#D7D7DD"}) \
    + p9.geom_hline(yintercept=2,linetype='dashed')\
    + p9.theme(
        figure_size=(8, 3),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),
    )
# plot.save(filename=results_path + "/zscore_density.pdf", dpi=300)
print(plot)
plot.save(filename=results_path + "/pval.pdf", dpi=300)



import plotnine as p9
from matplotlib import pyplot as plt
import pandas as pd
from sklearn.model_selection import train_test_split
import xgboost
from nanoCEM.cem_utils import extract_kmer_feature
from sklearn.metrics import roc_curve, auc
import numpy as np

plt.rcParams['pdf.fonttype'] = 42
results_path = 'f5c_result_rna_re'
df = pd.read_csv(results_path + '/current_feature.csv')

kmer_list = [3, 5, 7]
final_df = pd.DataFrame()

for item in kmer_list:
    feature_matrix, label = extract_kmer_feature(df, item, 2030)

    label[0] = label[0].apply(lambda x: 1 if x == 'Sample' else 0)

    X = feature_matrix.values
    y = label.values
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=2)
    model = xgboost.XGBClassifier()

    # Train the model on the training data
    model.fit(X_train, y_train)
    y_pred = model.predict_proba(X_test)[:, 1]

    # y_pred = model.predict_proba(X)[:,1]
    fpr, tpr, thresholds = roc_curve(y_test, y_pred)

    # 计算AUC（Area Under the Curve）
    roc_auc = auc(fpr, tpr)

    # 创建DataFrame
    df_roc = pd.DataFrame({'False Positive Rate': fpr, 'True Positive Rate': tpr})
    df_roc['kmer_length'] = str(item) + '-mer: ' + str(np.round(roc_auc, 2))
    final_df = pd.concat([final_df, df_roc], axis=0)

plot = p9.ggplot(final_df, p9.aes(x='False Positive Rate', y='True Positive Rate', color='kmer_length')) \
       + p9.scale_color_manual(['#9d906d', '#00c9aa', '#9f89ac']) \
       + p9.geom_smooth(method='loess', alpha=0.3) \
       + p9.geom_abline(linetype='dashed', alpha=0.8) \
       + p9.scale_y_continuous(breaks=[0, 0.25, 0.5, 0.75, 1],limits=(0, 1.15)) \
       + p9.labs(title='ROC Curve') \
       + p9.theme_bw() \
       + p9.theme(
    figure_size=(4, 4),
    panel_grid_minor=p9.element_blank(),
    axis_text=p9.element_text(size=13),
    axis_title=p9.element_text(size=13),
    title=p9.element_text(size=13, hjust=0.5),
    legend_position='bottom',
    legend_title=p9.element_blank(),
    strip_text=p9.element_text(size=13),
    strip_background=p9.element_rect(alpha=0),
)
plot.save(filename=results_path + "/roc_curve.pdf", dpi=300)
